
(* json extractor and printer *)


signature JSON = sig
    type location
    val toJson : Fiacre.Fiacre -> unit
    val getTransitions : string -> ((int * location list) list ref * (loc * location) list) option
    val recordTransition : ((int * location list) list ref * (loc * location) list) -> int * loc list -> unit
    val recordGtransition : string * string list * Fiacre.Interval -> unit
    val recordGPriorities : string list -> unit
    val printJson : unit -> unit 
end;


functor JsonFun (Prn : PRN) : JSON = struct

(* -- static part, generated by frac [-j | -json] : *)

open Fiacre;


(* static_json type *)

type typ = Fiacre.Type;

(* a json position is a pair (line,char)  [these are called location in front-fiacre ...]
    lines and chars are indexed from 0
    position (l,eol) is that of the  the last printable char in line l (eol encoded as ~1)
    position eof is the last position in file (encoded as (~1,~1))
*)

type position =
     {line : int, char : int};

(* a pair of positions identifying all characters between the first included and the second excluded *)
type location =
     {from : position, to : position};

type vardef = {
     sourcename: string,  (* toto *) 
     flatname: string,       (* comp_2_vtoto *)
     loc: location,
     vartype: typ
};

type statedef = {
     sourcename: string,  (* toto *) 
     flatname: string,       (* comp_2_vtoto *)
     loc: location
};

type typdef = {
     file : string,
     name: string, 
     loc: location,
     tdef: typ
};

(* local process transitions *)
type transition = {
     name : string,
     locations : location list     (* locations of elementary statements part of the transition *)
};

(* global transitions *)
type gtransition = {
    name : string,
    transitions : string list,      (* local transitions involved *)
    interval : Fiacre.Bound * Fiacre.Bound,
    lower : int list                    (* indices of gtransitions with lower priority *)
};

(* file, kind, flatname fields currently computed by to_json but not printed in json file *)
datatype component =
	 InstComponent of {
	 file : string,
	 kind : string,
	 name: string,
	 flatname: string,
	 loc: location,
	 inst: int,
	 vars: vardef list,
	 body: composition
         }
       | InstProcess of {
	 file : string,
	 kind : string,
	 name: string,
	 flatname: string,
	 loc: location,
	 inst: int,
	 vars: vardef list,
	 states: statedef list,
	 transitions: unit -> transition list  (* creation delayed at print time *)
	 }

and composition =
         Single of component
       | Comp of composition list;

type static_json = {
     arch_var : composition,                  (* one per fiacre section *)
     gtransitions: unit -> gtransition list,  (* global transitions (delayed at print time) *)
     deftyp :  typdef list,                   (* type definitions *)
     author : string,
     version : string,
     digits : unit -> int,
     flags : string list
};



(* --- extracting json *)

open Prn;

fun json_error m = (ErrPrint.print "Error: "; ErrPrint.print m; ErrPrint.print "\n"; ErrPrint.flush(); raise BYE);

(* returns next available integer for identifier id *)
val instance =
    let val created = ref ([] : (string * int) list)
	fun root ag = hd (String.fields (fn  #":" => true | _ => false) ag)
    in fn id => let val id = root id
		in case List.partition (fn (y,x) => y=id) (!created)  of
		       (nil,neg) => (created := (id,1) :: neg; 1)
		     | ((_,h)::_,neg) => (created := (id,h+1) :: neg; h+1)
		end
    end;

local
(* doubles any "_" in name to prevent ambiguity of instantiated names: *)
 fun encode id = String.translate (fn #"_" => "__" | c => Char.toString c) id;
in
(* return aliased variables and states *)
fun compName name inst = encode name ^ "_" ^ Int.toString inst;
fun varName proc var inst = encode proc ^ "_" ^ Int.toString inst ^ "_v" ^ encode var;
fun stateName proc state inst = encode proc ^ "_" ^ Int.toString inst ^ "_s" ^ encode state;
fun transName proc tr inst = encode proc ^ "_" ^ Int.toString inst ^ "_t" ^ (Int.toString tr);
end;


(* computing and recording elementary statements in processes: *)

(* recording statement locations and transitions for proc *)

local
(* for each proc, maps the position identifying an elementary statement to its actual location in source *)
val STATEMENTS = ref ([] : (string * (loc * location) list) list);

(* stores process transitions (reversed), for each proc *)
val TRANSITIONS = ref ([] :  (string * (int * location list) list ref) list);

(* stores global transitions (reversed) *)
val GTRANSITIONS = ref ([] :  (string * string list * Fiacre.Interval) list);
val GPRIORITIES = ref ([] :  string list list);

in
(* storing elementary statement locations *)

(* prints STATEMENTS on stderr, for debug *)
fun pStatements () =
    let open ErrPrint
	fun ploc (l,c) = (printi l; print "."; printi c)
	fun plocation l =
	    (print "from="; ploc(#line (#from l),#char (#from l)); print ", to ="; ploc(#line (#from l),#char (#from l)))
    in List.app (fn (proc,st) =>
		    (print "\nprocess = "; print proc; print "\n";
		     print "statements = "; 
		     print_list ("\n","\n","\n") (fn (id,frto) => (print "id = "; ploc id; print ", location : "; plocation  frto))  st)) (!STATEMENTS)
    end;

(* records the elementary statements of process proc *)
(* ### note: statements i init clause not recorded, should they be ? ### *)
fun recordStatements trl tends proc inst =
    let val ELEMST = ref [];
	fun mkloc ((l1,c1),(l2,c2)) = {from = {line=l1, char=c1}, to = {line=l2, char=c2}}
	fun extend id loc = ELEMST := (id,mkloc loc) :: !ELEMST;
	infix ++
	fun (l,c) ++ x = (l,c+x)
	fun record left s right = (* left is left bound, right is right bound *)
	    case s of
		(* elementary statements *)
		StNull loc => extend loc (loc, (fst loc, snd loc +3))
	      | StTo (loc,state) => extend loc (loc, right)
	      | StAssign (loc,patt,e,_) => extend loc (left, right)
	      | StAny (loc,_,_,_) => extend loc (left, right)
	      | StSignal ((loc,port),_,_) => extend loc (loc, (#1 loc, #2 loc + String.size port))
	      | StInput ((loc,_),_,_,_,_) => extend loc (loc, right)
	      | StOutput ((loc,_),_,_,_) => extend loc (loc, right)
	      | StOn (loc,e) => extend loc (loc, right)
	      | StWait (loc,b1,b2) => extend loc (loc, right)
	      | StTag (loc,tag) => extend loc (loc, right)
	      | StReturn _ => () (* never found in a process transition *)
	      (* compound statements *)
	      | StForeach (loc,_,_,lft,s,rht) => record (lft ++ 2) s rht
	      | StWhile (loc,c,lft,s,rht) => record (lft ++ 2) s rht
	      | StCond (loc,c,(tloc,s1),lcsl,s2,eloc) =>
		((* locating then statement: *)
		 case lcsl of
		     (lloc,cc,ss)::_ => record (tloc ++ 4) s1 lloc
		   | _ => record (tloc ++ 4) s1 (case s2 of SOME (elloc,_) => elloc | _ => eloc);
		 (* locating elsif statements *)
		 let val ends =
			 let val l = List.map (fn (loc,_,_) => loc) lcsl
			 in (tl l @ [case s2 of SOME (elloc,_) => elloc | _ => eloc]) handle _ => []
			 end
		 in List.app (fn ((lloc,cc,ss),lend) => record (lloc ++ 5) ss lend) (ListPair.zip (lcsl, ends))
		 end;
		 (* locating else statement *)
		 case s2 of SOME (elloc,s) => record (elloc ++ 4) s eloc | _ => ())
	      | StCase (loc,_,ml,_,eloc) => (* List.app (fn (loc,_,_,_,s) => record loc s right) ml *)
		let val ends = tl (List.map (fn (loc,_,_,_,_) => loc) ml) @ [eloc]
		in List.app (fn ((_,_,_,stloc,s),eloc) => record (stloc ++ 2) s eloc) (ListPair.zipEq (ml,ends))
		end
	      | StSelect (sl,endl) =>
		let val fsll = List.concat sl
		    val ends = tl (List.map fst fsll) @ [endl]
		    val szs = (* size of heads; select and unless are of same length :-) *)
			flat (List.map (fn g => List.tabulate (List.length g, fn 0 => 6 | _ => 2)) sl)
		in List.app (fn ((left,s),(sz,right)) => record (left ++ sz) s right) (ListPair.zipEq (fsll,ListPair.zip (szs,ends)))
		end
	      | StSequence (loc,s1,s2) => (record left s1 loc; record loc s2 right)
    in if inst > 1 then () else
       (List.app (fn ((tloc,(sloc,src_state),s),to) =>
		     (extend tloc (tloc, sloc ++ String.size src_state);
		      record (sloc ++ String.size src_state) s to)) (ListPair.zipEq (trl,tends));
	STATEMENTS := (proc,rev (!ELEMST)) :: !STATEMENTS)
    end;


(* storing transition locations of processes *)

(* returns handles for adding transitions for id, as an option *)
fun getTransitions id =
    ((assoc id (!TRANSITIONS); NONE)
     handle _ => let val r = ref [] in TRANSITIONS := (id,r) :: !TRANSITIONS; SOME (r,assoc id (!STATEMENTS)) end)
    handle x =>
	   (*
	    let open ErrPrint
	    in print "id = "; print id; print ", heads = "; print_list ("[",", ","]\n") (fn (s,_) => print s) (!STATEMENTS); raise x
	    end *)
	   (* no statements have been recorded for id; process id is declared (hence getTransitions called),
	      but is never called from main (hence recordTransition never called for that id) *)
	   NONE;

(* only called if newTransitions id returned true *)
fun recordTransition (r,sr) (n,tr) =
    r := (n, List.map (fn t => assoc t sr) tr) :: !r;

(* for toJson: *)
fun extractTransitions proc inst =
    let fun mktrans (no,trl) = {name=transName proc no inst, locations=trl}
    in List.map mktrans (rev (!(assoc proc (!TRANSITIONS))))
    end;


(* storing global transitions *)

fun scaleBound b =
    Real.floor (b * (pow (10.0,!DIGITS)));

(* records transition names, components, intervals *)
fun recordGtransition tr =
     GTRANSITIONS := tr :: !GTRANSITIONS;

(* records transition priorities *)
fun recordGPriorities prl =
    GPRIORITIES := prl :: !GPRIORITIES;

fun extractGtransitions () =
    let val rg = rev (ListPair.zipEq (!GTRANSITIONS,!GPRIORITIES))
	(* index of x in list l (from 0) [fails with Match if not member] *)
	fun index x =
	    let fun rindex ((h,_)::t) n = if #1 h=x then n else rindex t (n+1)
	    in rindex rg 0
	    end
    in List.map (fn ((k,b,i),prl) => {name=k, transitions=b, interval=i, lower=List.map index prl}) rg
    end;

end;

(* start position of a declaration (approximate) *)
fun declPos d =
     case d of ProcessDecl x => #1 x  (* position of name - length of "process" - 1 *)
	     | ComponentDecl x => #1 x
	     | TypeDecl x => #1 x
	     | ChannelDecl x => #1 x
	     | ConstantDecl x => #1 x
	     | FunctionDecl x => #1 x
	     | ExternDecl x => #1 x
	     | Main x => #1 (#1 x)
	     | PropertyDecl x => #1 x
	     | Assert x => #1 x;


(* extracts json rep from program *)
fun to_json (Program dl) =
    let val file = !Input.name  (* single file, currently *)
	(* declarations associated with an end position (approximate) *)
	val starts = List.map declPos dl
	val ends = ListPair.zipEq (dl, tl starts @ [(~1,~1)])  (* (~1,~1) encodes end-of-file *) 
	fun lookup n = case List.find (fn (ProcessDecl (loc,(_,name),pl,al,sl,vl,init,trl),_) => name = n
					| (ComponentDecl (loc,(_,name),pl,al,vl,lpl,prio,init,comp),_) => name = n
					| _ => false) ends of
			   SOME d => d
	fun flatten l =  flat (List.map (fn (idl,a,b) => List.map (fn v => (v,a,b)) idl) l)
	fun append l1 l2 = (* (l1 - hidden by l2) @ l2 *)
	    let val v2 = map #sourcename l2 in List.filter (fn x => not (mem (#sourcename x) v2)) l1 @ l2 end
	fun mkpos (l,c) = {line=l, char=c}
	fun mkloc (l,c) v = {from = mkpos (l,c), to = mkpos (l,c + String.size v)}
	(* shared variables are named from the component in which they were defined.
	    shenv : local shared var name -> C shared var name
	    shvmap : built by instance caller. shvmap[i] = NONE if args[i] not a ref | SOME v if args[i] is an alias for C shared var v *)
	fun getArch shenv shvmap (ProcessDecl (locp,(loc,id),pl,al,sl,vl,init,trl),to) =
	    let val fal = flatten al
		val shenv = (* extend shenv according to shvmap *)
		    List.mapPartial (fn (SOME shv,(RefArg (_,v),_,_)) => SOME (v,shv) | _ => NONE) (ListPair.zipEq (shvmap,fal)) @ shenv
		val i = instance id
		val avars = List.map (fn (ValArg (loc,v),_,ty) => {sourcename=v,flatname=varName id v i, loc=mkloc loc v,vartype=ty}
				       | (RefArg (loc,v),_,ty) => {sourcename=v,flatname=assoc v shenv, loc=mkloc loc v,vartype=ty}) (flatten al)
		val lvars = List.map (fn ((loc,v),ty,_) => {sourcename=v,flatname=varName id v i, loc=mkloc loc v,vartype=ty}) (flatten vl)
		val tends = tl (List.map (fn (loc,_,_) => loc) trl) @ [to]
	    in (* let open ErrPrint in print "recording statements of "; print id; print "\n" end; *)
               recordStatements trl tends id i;
	       InstProcess {file=file,
			    kind="process",
			    name = id, 
			    flatname = compName id i,
			    (* loc = mkloc loc id (* locates the name of the process in declaration *) *)
			    loc = {from = mkpos locp, to = mkpos to}, (* locates the full process declaration *)
			    inst = i,
			    vars=append avars lvars,
			    states = List.map (fn (loc,s) => {sourcename=s, flatname=stateName id s i, loc=mkloc loc s}) sl,
			    transitions = (fn () => extractTransitions id i)}
	    end
	  | getArch shenv shvmap (ComponentDecl (locc,(loc,id),pl,al,vl,lpl,prio,init,comp),to) =
	    let val fal = flatten al
		val fvl = flatten vl
		val shenv = (* extend shenv according to shvmap *)
		    List.mapPartial (fn (SOME shv,(RefArg (_,v),_,_)) => SOME (v,shv) | _ => NONE) (ListPair.zipEq (shvmap,fal)) @ shenv
		val i = instance id
		val avars = List.map (fn (ValArg (loc,v),_,ty) => {sourcename=v,flatname=varName id v i, loc=mkloc loc v,vartype=ty}
				       | (RefArg (loc,v),_,ty) =>  {sourcename=v,flatname=assoc v shenv, loc=mkloc loc v,vartype=ty}) (fal)
		val lvars = List.map (fn ((loc,v),ty,_) => {sourcename=v,flatname=varName id v i, loc=mkloc loc v,vartype=ty}) fvl
		val shenv = (* extend shenv with locally defined variables [local var(s) may hide shared var(s)] *)
		    List.map (fn ((_,v),_,_) => (v,varName id v i)) fvl @ shenv
	    in InstComponent {file=file,
			      kind="component",
			      name=id,
			      flatname = compName id i,
			      (* loc = mkloc loc id (* locates the name of the component in declaration *) *)
			      loc = {from = mkpos  locc, to = mkpos to}, (* locates the full component declaration *)
			      inst=i,
			      vars=append avars lvars,
			      body=getComp shenv comp}
	    end
	and getComp shenv c =
	    case c of
		InstanceComp ((loc,id),_,el) =>
		let val shvmap = List.map (fn RefExp (v,_) => SOME (assoc v shenv) | _ => NONE) el
		in Single (getArch shenv shvmap (lookup id))
		end
	      | ParComp (loc,_,pcl) => Comp (List.map (getComp shenv o snd) pcl)
	val components =
	    (* builds a composition from the mains of the different sections *)
	    let val mains = List.mapPartial (fn Main ((loc,id),_) => SOME id | _ => NONE) dl
	    in (* let open ErrPrint in print "mains = "; print_list ("[",", ","]\n") print mains end; *)
		case mains of
		   []   => json_error "missing main, cannot emit json description"
		 | [c] => (* single section *) Single (getArch [] [] (lookup c))
		 | cl  => (* several section, each a single composition *)
		   Comp (List.map (Single o getArch [] [] o lookup) cl)
	    end
	val tydecls = List.mapPartial (fn TypeDecl (_,(loc,id),body) => SOME {file=file,name=id, loc=mkloc loc id, tdef=body} | _ => NONE) dl
    in {deftyp = tydecls,
	arch_var = components,
	gtransitions = (fn () => extractGtransitions ()),
	author = "Created " ^ Date.toString (Date.fromTimeLocal(Time.now ())) ^ " by FRAC version " ^ version,
	version = "1",
	digits = fn _ => !DIGITS,
	flags = List.filter (fn s => #"-" = String.sub(s,0)) (CommandLine.arguments())
       }
    end;





(* -- printing json *)

(* EXPVAL defined in front-fiacre; filled by typer, needed to compute lengths in queues and arrays *)
fun intVal e = let val IntExp (len,_) = !EXPVAL e in len end;

fun print_string s =
    (print "\""; print s; print "\"");
fun print_member (tag,value,p) =
    (print_string tag; print " : "; p value);
fun print_array p a =
    print_list ("[",", ","]") p a;

(* prints json representation of a static_json value *)
fun print_json (j : static_json) =
    (print "{";
     print_member ("arch_var",#arch_var j,print_composition); print ", ";
     print_member ("deftyp",#deftyp j,print_array print_deftyp); print ", ";
     print_member ("gtransitions",#gtransitions j (),print_array print_gtransition); print ", ";
     print_member ("author",#author j,print_string); print ", ";
     print_member ("version",#version j,print_string); print ", ";
     print_member ("digits",#digits j (),printi); print ", ";
     print_member ("flags",#flags j,print_array print_string);     
     print "}")
and print_component c =
    case c of
	 InstComponent r => 
	 (print "{";
	  (* print_member ("file", #file r, print_string); print ", "; *)
	  (* print_member ("kind", #kind r, print_string); print ", "; *)
	  print_member ("name", #name r, print_string); print ", ";
	  print_member ("flatname", #flatname r, print_string); print ", ";
	  print_member ("loc", #loc r, print_location); print ", ";
	  print_member ("inst",#inst r,printi); print ", ";
	  print_member ("vars",#vars r,print_array print_vardef); print ", ";
	  print_member ("body",#body r, print_composition);
	  print "}")
       | InstProcess r =>
	 (print "{";
	  (* print_member ("file", #file r, print_string); print ", "; *)
	  (* print_member ("kind", #kind r, print_string); print ", "; *)
	  print_member ("name",#name r,print_string); print ", ";
	  print_member ("flatname", #flatname r, print_string); print ", ";
	  print_member ("loc",#loc r,print_location); print ", ";
	  print_member ("inst",#inst r,printi); print ", ";
	  print_member ("vars",#vars r,print_array print_vardef); print ", ";
	  print_member ("states",#states r,print_array print_statedef); print ", ";
	  print_member ("transitions",#transitions r (),print_array print_transition);
          (* transitions unknown when toJson is called; their computation is delayed at print time *)
	  print "}")
and print_composition c =
    case c of
	Single c => print_component c
      | Comp cl => print_array print_composition cl
and print_vardef r =
    (print "{";
     print_member ("sourcename",#sourcename r,print_string); print ", ";
     print_member ("flatname",#flatname r,print_string); print ", ";
     print_member ("loc",#loc r,print_location); print ", ";
     print_member ("vartype",#vartype r,print_type);
    print "}")
and print_statedef r =
    (print "{";
     print_member ("sourcename",#sourcename r,print_string); print ", ";
     print_member ("flatname",#flatname r,print_string); print ", ";
     print_member ("loc",#loc r,print_location);
     print "}")
and print_type ty =
    case ty of
	BoolType => print_string "bool"
      | NatType => print_string "nat"
      | IntType => print_string "int"
      | NamedType name => print_string name
      | IntervalType (IntExp (from,_),IntExp (to,_)) =>
	(print "{";
	 print_member ("interval",
		       [("low",from),("high",to)],
		       print_list ("{",",","}") (fn (f,v) => (print_string f; print ": "; printi v)));
	 print "}")
       | UnionType ctyl =>
	(print "{";
	 print_member ("union",
		       List.concat (List.map (fn ([c],ty) => [(c,ty)] | (l,ty) => List.map (fn c => (c,ty)) l) ctyl),
		       print_list ("{",", ","}") (fn (c,BotType) => print_member (c,[],print_array print_type)
						  | (c,ty) => print_member (c,[ty],print_array print_type)));
	 print "}")
      | RecordType ftyl =>
	(print "{";
	 print_member ("record",
		       List.concat (List.map (fn ([f],ty) => [(f,ty)] | (l,ty) => List.map (fn f => (f,ty)) l) ftyl),
		       print_list ("{",", ","}") (fn (f,ty) => print_member (f,ty,print_type)));
	 print "}")
       | ArrayType (e,ty) =>
	 (* working on sources; length may be an expression, and may include constants ... *)
	 let val len = intVal e
	 in print "{";
	    print_member ("array",(len,ty),(fn (len,ty) => (print "{";
							    print_member ("len",len,printi); print ", ";
							    print_member ("elemtype",ty,print_type);
							    print "}")));
	    print "}"
	 end
       | QueueType (e,ty) =>
	 (* working on sources; length may be an expression, and may include constants ... *)
	 let val len = intVal e
	 in print "{";
	    print_member ("queue",(len,ty),(fn (len,ty) => (print "{";
							    print_member ("len",len,printi); print ", ";
							    print_member ("elemtype",ty,print_type);
							    print "}")));
	    print "}"
	 end
and print_deftyp dty =
    (print "{";
     (* print_member ("file",#file dty,print_string); print ", "; *)
     print_member ("name",#name dty,print_string); print ", ";
     print_member ("loc",#loc dty,print_location); print ", ";
     print_member ("tdef",#tdef dty,print_type);
     print "}")
and print_position e =
     (print "{";
      print_member ("line", #line e, printi); print ", ";
      print_member ("char", #char e, printi);
      print "}")
and print_location p =
    (print "{";
      print_member ("from", #from p, print_position); print ", ";
      print_member ("to", #to p, print_position);
      print "}")
and print_transition p =
    (print "{";
     print_member ("name", #name p, print_string); print ", ";
     print_member ("locations", #locations p, print_array print_location);     
     print "}")
and print_gtransition p =
    (print "{";
     print_member ("name", #name p, print_string); print ", ";
     print_member ("transitions", #transitions p, print_array print_string);  print ",";
     print_member ("interval", #interval p, fn (l,r) => (print "{";
							 print_member ("left",l,print_bound); print ",";
							 print_member ("right",r,print_bound);
							 print "}")); print ",";
     print_member ("lower", #lower p, print_array printi);
     print "}")
and print_bound b =
    case b of
      Infinite => print_string "infinite"
    | Open x => (print "{"; print_member ("open",scaleBound x,printi); print "}")
    | Closed x => (print "{"; print_member ("closed",scaleBound x,printi); print "}")
;




(* main routines *)

val JSONDESC = ref NONE;

fun toJson p =
    if HPX andalso !HIPPOEXT then () else (JSONDESC := SOME (to_json p) handle _ => failwith "toJson");

fun printJson () =
    if HPX andalso !HIPPOEXT then () else (case !JSONDESC of NONE => () | SOME j => (print_json j (*  ; pStatements ()  *)));


(* -- dynamic part (function sprint_json_state in generated C)

type dynamic_json = valvar list
type valvar = flatname * value
datatype value =  
	  BoolExp of bool
	| IntExp of int * Type
	| RecordExp of (string * Exp) list * Type
	| ArrayExp of Exp list * Type
	| QueueExp of Exp list * Type
	| ConExp of string * Type                           (* IdentExp (_,(ConstrKind,ty)) when ty is a general union *)
	| ConstrExp of string * Exp * Type             (* construction, JuxtExp(IdentExp (_,(ConstrKind,ty)),...) *)
*)

end; (* JsonFun *)


structure Json = JsonFun(OutPrint);

